#!/usr/bin/python3

import re  # NOQA
from collections import OrderedDict
from jirate.decor import pretty_date, vsep_print, comma_separated
from jirate.jira_custom import custom_field_renderers


#
# Field rendering functions. Return a string, or None if you want the field
# suppressed.
#
def _list_of_key(field, key, as_object=False):
    if as_object:
        return list([item[key] for item in field])
    return comma_separated([item[key] for item in field])


def string(field, fields, as_object=False):
    return field


def floating(field, fields, as_object=False):
    if str(field).endswith('.0'):
        field = int(field)
    if not as_object:
        field = str(field)
    return field


def auto_field(field, fields, as_object=False):
    if isinstance(field, str):
        return field
    elif isinstance(field, list):
        return array(field, fields, as_object)
    elif isinstance(field, dict):
        for key in ['name', 'value']:
            if key in field:
                if 'id' in field and not as_object:
                    return f"{field[key]} (ID: {field['id']})"
                return auto_field(field[key], fields, as_object)
    elif isinstance(field, float):
        return floating(field, fields, as_object)

    if as_object:
        return field
    else:
        return str(field)


def option_with_child(field, fields, as_object=False):
    if 'child' in field:
        return field['value'] + ' - ' + field['child']['value']
    return field['value']


def key(field, fields, as_object=False):
    return field['key']


def value(field, fields, as_object=False):
    return field['value']


def name(field, fields, as_object=False):
    return field['name']


def user(field, fields, as_object=False):
    return field['displayName'] + ' - ' + field['emailAddress']


def user_list(field, fields, as_object=False):
    # Until we take email addresses as input, we should use
    # user names only when presenting lists of users
    return _list_of_key(field, 'name', as_object)


def array(field, fields, as_object=False):
    if as_object:
        return list(field)
    return comma_separated([str(item) for item in field])


def value_list(field, fields, as_object=False):
    return _list_of_key(field, 'value', as_object)


def name_list(field, fields, as_object=False):
    return _list_of_key(field, 'name', as_object)


def date(field, fields, as_object=False):
    return pretty_date(field)


def datetime(field, fields, as_object=False):
    return pretty_date(field)


def _ratio(field, fields, as_object=False):
    if int(field) < 0:
        return None
    if as_object:
        return field
    return str(field)


def _priority(field, fields, as_object=False):
    if field['name'] not in ('Undefined', 'undefined'):
        return field['name']
    return None


def _reporter(field, fields, as_object=False):
    if field['emailAddress'] != fields['creator']['emailAddress']:
        return user(field, fields)
    return None


def _created_updated(field, fields, as_object=False):
    created = pretty_date(field)
    if field != fields['updated']:
        updated = pretty_date(fields['updated'])
        return f'{created} (Updated {updated})'
    return created


def _human_list(basic_value, item_list, field, fields, as_object=False):
    if field[basic_value] in (0, '0'):
        return None
    if item_list not in field or as_object:
        return auto_field(field[basic_value], fields, as_object)
    else:
        # Special: votes is a count + list_of_key as displayName
        return str(field[basic_value]) + ': ' + _list_of_key(field[item_list], 'displayName', False)


def _votes(field, fields, as_object=False):
    return _human_list('votes', 'voters', field, fields, as_object)


def _watchers(field, fields, as_object=False):
    return _human_list('watchCount', 'watchers', field, fields, as_object)


# these can functions can be referenced in custom user
# config files.  These should be generic and should not
# utilize custom fields.
_field_renderers = {
    'string': string,
    'any': auto_field,
    'number': auto_field,
    'option': value,
    'option-with-child': option_with_child,
    'key': key,
    'issuelinks': key,
    'value': value,
    'name': name,
    'version': name,
    'user': user,
    'user_list': user_list,
    'array': array,
    'email_list': user_list,  # Is this duplicate needed?
    'value_list': value_list,
    'name_list': name_list,
    'datetime': datetime,
    'date': date
}


# Rule of thumb: if it's using an above generic
# renderer, use the key above.  If it's custom
# for that field, use the function. The goal is to
# provide as many generic items as possible to reduce
# the need for config-side rendering via code snippets.
#
# That said, in the config, one can call the functions:
#
#    "code": "name(field, fields)"
#
#     ===
#
#    "display": "name"
#
# 'id': matches your JIRA instance
# 'name': What to display as output when printing
# 'immutable': If True, user configurations cannot
#              overwrite/register handler for this
#              field
# 'display': Omitted: Display the field rendered as a
#                     string
#            False: Do not display this field at all
#            String: Predefined key from above field
#                    renderer table.
# 'code': Should really only be used by custom user
#         configs when here_there_be_dragons is True
#         Note: 'display' supersedes 'code' if both
#         are present.
# 'verbose': if True, Only show this field in
#            verbose mode when rendering all fields
#
_base_fields = [
    {
        'id': 'issuetype',
        'name': 'Issue Type',
        'immutable': True,
        'display': 'name'
    },
    {
        'id': 'parent',
        'name': 'Parent',
        'immutable': True,
        'display': 'key'
    },
    {
        'id': 'priority',
        'name': 'Priority',
        'display': _priority
    },
    {
        'id': 'created',
        'name': 'Created',
        'immutable': True,
        'verbose': True,
        'display': _created_updated
    },
    {
        'id': 'duedate',
        'name': 'Due Date',
        'display': 'date'
    },
    {
        # This is part of above.
        'id': 'updated',
        'name': 'Updated',
        'immutable': True,
        'display': False
    },
    {
        'id': 'assignee',
        'name': 'Assignee',
        'display': 'user'
    },
    {
        'id': 'status',
        'name': 'Status',
        'display': 'name'
    },
    {
        'id': 'resolution',
        'name': 'Resolution',
        'display': 'name'
    },
    {
        'id': 'resolutiondate',
        'name': 'Resolved',
        'display': 'date'
    },
    {
        'id': 'security',
        'name': 'Security Level',
        'display': 'name'
    },
    {
        'id': 'workratio',
        'name': 'Work Ratio',
        'display': _ratio
    },
    {
        'id': 'creator',
        'name': 'Creator',
        'display': 'user',
        'verbose': True,
    },
    {
        'id': 'reporter',
        'name': 'Reporter',
        'verbose': True,
        'display': _reporter
    },
    {
        'id': 'archivedby',
        'name': 'Archiver',
        'verbose': True,
        'display': 'user'
    },
    {
        'id': 'archiveddate',
        'name': 'Archived',
        'verbose': True,
        'display': 'date'
    },
    {
        'id': 'labels',
        'name': 'Labels',
        'display': 'array',
    },
    {
        'id': 'components',
        'name': 'Component(s)',
        'display': 'name_list'
    },
    {
        'id': 'versions',
        'name': 'Affects Version(s)',
        'display': 'name_list'
    },
    {
        'id': 'fixVersions',
        'name': 'Fix Version(s)',
        'display': 'name_list'
    },
    {
        'id': 'votes',
        'name': 'Votes',
        'display': _votes
    },
    {
        'id': 'watches',
        'name': 'Watchers',
        'display': _watchers
    },
]


# Shortcut to avoid making above list enormous
# These are all standard fields we don't presently
# have rendering for, but might in the future.
# Remember ordering matters in above list.  Since
# it "shouldn't" for things we're not printing out,
# alphabetic is fine.
_quiet_fields = [
    'aggregateprogress',
    'aggregatetimeestimate',
    'aggregatetimeoriginalestimate',
    'aggregatetimespent',
    'archivedby',
    'archiveddate',
    'environment',
    'lastViewed',
    'progress',
    'project',
    'reporter',
    'timeestimate',
    'timeoriginalestimate',
    'timespent',
    'timetracking',
    'worklog',
    'workratio'
]

_quiet_fields_applied = False


if not _quiet_fields_applied:
    _quiet_fields_applied = True
    for field in _quiet_fields:
        val = {'id': field, 'name': field, 'display': False}
        _base_fields.append(val)


# These are fields we should never provide
# custom rendering for; they are inherently
# complicated or positional.
_ignore_fields = [
    'attachment',
    'comment',
    'description',
    'issuekey',
    'issuelinks',
    'subtasks',
    'summary',
    'thumbnail'
]


_fields = None


_array_renderers = {
    'user': 'user_list',
    'option': 'value_list',
    'version': 'name_list',
    'group': 'name_list'
}


_jirate_fields = {}

# Store modules here so we don't keep loading from disk
_loaded_mods = {}


def apply_schema_renderer(field):
    schema = field['schema']
    if 'custom' in schema and schema['custom'] in custom_field_renderers:
        field['display'] = custom_field_renderers[schema['custom']]
        return
    if schema['type'] == 'array':
        try:
            field['display'] = _array_renderers[schema['items']]
        except KeyError:
            field['display'] = array
    else:
        try:
            field['display'] = _field_renderers[schema['type']]
        except KeyError:
            field['display'] = string


def func_from_path(filename, function, field, fields):
    # Load a function from a file and run that
    import importlib.util
    import os

    full_fn = os.path.expanduser(filename)
    mod_fn = os.path.basename(full_fn)
    if mod_fn not in _loaded_mods:
        spec = importlib.util.spec_from_file_location(f'{mod_fn}', full_fn)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        _loaded_mods[mod_fn] = module
    else:
        module = _loaded_mods[mod_fn]

    ret = eval(f'module.{function}(field, fields)')
    return ret


def eval_custom_field(__code__, field, fields):
    """Proof of concept: Execute inline code to render a field

    Only used if 'here_there_be_dragons' is set to true.  Represents
    an obvious security issue if you are not in control of your
    jirate configuration file:
        "code": "os.system('rm -rf ~/*')"

    Parameters:
      field:    is your variable name for your dict
      fields:   dict of fields indexed by id
      __code__: is inline in your config and can reference field

    Returns:
      rendered field value (string)
    """
    if field is None or not field:
        return None
    if '__code__' in __code__:
        raise ValueError('Reserved keyword in code snippet')
    if __code__.startswith('#'):
        filename, func = __code__[1:].split(':')
        return func_from_path(filename, func, field, fields)
    return eval(str(__code__))


def apply_field_renderers(custom_field_defs=None, reorder_custom=True):
    """Custom field rendering setup function

    Parameters:
      custom_field_defs: Dictionary (typically retrieved from
        /rest/api/latest/field) with custom code snippets or
        field rendering definitions

    Returns:
      nothing in particular
    """
    global _fields
    base_fields = OrderedDict()
    custom_fields = OrderedDict()
    ret = OrderedDict()

    # NOTE: Don't feed in field definitions acquired from getting
    # createmeta or editmeta; the resulting dictionary is subtly
    # different in an important way: field['id'] is field['fieldId']
    if not custom_field_defs:
        for field in _base_fields:
            if field['id'] in _ignore_fields:
                continue
            ret[field['id']] = field
        _fields = ret
        return

    # First go through base fields
    for field in _base_fields:
        base_fields[field['id']] = field

    # reorder
    for field in custom_field_defs:
        if 'id' not in field:
            continue
        if field['id'] in _ignore_fields:
            continue
        if field['id'] in base_fields:
            bf = base_fields[field['id']]
            if 'immutable' in bf and bf['immutable']:
                custom_fields[field['id']] = bf
                continue
            custom_fields[field['id']] = field
            if 'display' in bf and 'display' not in field and 'code' not in field:
                custom_fields[field['id']]['display'] = bf['display']
                continue
        custom_fields[field['id']] = field
        if 'display' not in field and 'code' not in field and 'schema' in field:
            apply_schema_renderer(field)
        if '_jirate_reference' in field:
            # Just keep track of reference
            _jirate_fields[field['id']] = field['_jirate_reference']

    if reorder_custom:
        for key in base_fields:
            if key not in custom_fields:
                ret[key] = base_fields[key]
        for key in custom_fields:
            ret[key] = custom_fields[key]
    else:
        for key in base_fields:
            ret[key] = base_fields[key]
        for key in custom_fields:
            ret[key] = custom_fields[key]

    if _fields:
        # If we were called twice, tack on the things we'd already set up
        for key in _fields:
            if key not in ret:
                ret[key] = _fields[key]

    _fields = ret


def jirate_field(field_key):
    if field_key in _jirate_fields:
        return _jirate_fields[field_key]
    return None


def render_field_data(field_key, fields, verbose=False, allow_code=False, as_object=False):
    """Render the field using custom-renderers or user-supplied code
    Note: you must first configure the rendering engine using apply_field_renderers()

    Parameters:
      field_key: Key of field in raw JIRA JSON we're rendering
      fields: The entire raw JIRA JSON we're using
      verbose: Passed to field rendering function
      allow_code: Allow eval() during execution (dangerous whenever source
                  is untrusted)
      as_object: Return a Python object instead of a string representation

    Returns:
      field_name: Human-readable field name (string)
      value: Rendered field value (string)
    """
    if field_key not in _fields:
        return field_key, fields[field_key]
    field_name = _fields[field_key]['name']
    if field_key not in fields and field_key not in _jirate_fields:
        return field_name, None
    if field_key in _jirate_fields:
        real_key = _jirate_fields[field_key]
        if real_key not in fields:
            return field_name, None
        field = fields[_jirate_fields[field_key]]
    else:
        field = fields[field_key]
    if not field:
        return field_name, None
    field_config = _fields[field_key]

    if 'verbose' in field_config:
        if field_config['verbose'] is True and not verbose:
            return field_name, None
    if 'disabled' in field_config and field_config['disabled'] is True:
        return field_name, None
    # display supersedes code
    if 'display' in field_config:
        r_info = field_config['display']
        if isinstance(r_info, bool):
            if not r_info:
                return field_name, None
            else:
                if as_object:
                    return field_name, field
                return field_name, str(field)
        if isinstance(r_info, str):
            if r_info not in _field_renderers:
                return field_name, f'<invalid renderer: {r_info} for {field_key}>'
            else:
                try:
                    ret = _field_renderers[r_info](field, fields, as_object)
                except Exception as exc:
                    ret = '<' + str(exc) + '>'
                return field_name, ret
        else:
            return field_name, r_info(field, fields, as_object)
    if 'code' in field_config and allow_code:
        try:
            ret = eval_custom_field(field_config['code'], field, fields)
        except Exception as exc:
            ret = '<' + str(exc) + '>'
        return field_name, ret
    if as_object:
        return field_name, field
    return field_name, str(field)


def field_ordering():
    return [_fields.keys()]


def max_field_width(issue, verbose, allow_code):
    width = 0

    for field_key in _fields:
        field_name, val = render_field_data(field_key, issue, verbose, allow_code)
        if not val:
            continue
        width = max(width, len(field_name))
    return width


def render_issue_fields(issue, verbose=False, allow_code=False, width=None):
    if not width:
        width = max_field_width(issue, verbose, allow_code)

    for field_key in _fields:
        field_name, val = render_field_data(field_key, issue, verbose, allow_code)
        if not val:
            continue
        vsep_print(' ', 0, field_name, width, val)
        # print(field_name.ljust(width), sep, val)



#!/usr/bin/env python

from jirate.jboard import Jirate
from jirate.tests import fake_jira, fake_user, fake_transitions

import pytest  # NOQA
import types


fake_jirate = Jirate(fake_jira())


def transitions_override(obj, issue):
    return fake_transitions['transitions']


# XXX we don't have subs in at the level we need, so fake transitions this way
fake_jirate.transitions = types.MethodType(transitions_override, fake_jirate)


def test_jirate_myself():
    # assert not fake_jirate._user
    me = fake_jirate.user
    assert fake_jirate.user
    assert me == fake_user


def test_jirate_issue():
    issue = fake_jirate.issue('TEST-9999')
    assert issue is None
    issue = fake_jirate.issue('TEST-1')
    assert issue


def test_jirate_basicfields():
    # Normal fields should be accessible via the field() function
    issue = fake_jirate.issue('TEST-2')

    val = issue.fields.description

    assert fake_jirate.field(issue, 'description') == val
    assert issue.field('description') == val


def test_jirate_customfields():
    issue = fake_jirate.issue('TEST-1')

    # change if you alter issue or above predefined fake issue
    # data
    expected = 'test-build-1'

    assert issue.fields.customfield_1234567 == expected
    assert issue.raw['fields']['customfield_1234567'] == expected

    # too bad...
    with pytest.raises(AttributeError):
        issue.fields.fixed_in_build

    # Test one-way
    assert fake_jirate.field('TEST-1', 'fixed_in_build') == expected
    assert fake_jirate.field('TEST-1', 'Fixed in Build') == expected
    assert fake_jirate.field('TEST-1', 'customfield_1234567') == expected

    # Negative test
    with pytest.raises(AttributeError):
        fake_jirate.field('TEST-2', 'arglebargle')

    assert fake_jirate.field(issue, 'fixed_in_build') == expected
    assert fake_jirate.field(issue, 'Fixed in Build') == expected
    assert fake_jirate.field(issue, 'customfield_1234567') == expected

    # Negative test
    with pytest.raises(AttributeError):
        fake_jirate.field(issue, 'arglebargle')

    # Test issue
    assert issue.field('fixed_in_build') == expected
    assert issue.field('Fixed in Build') == expected
    assert issue.field('customfield_1234567') == expected

    # Negative test
    with pytest.raises(AttributeError):
        issue.field('arglebargle')


def test_transition_resolutions():
    issue = fake_jirate.issue('TEST-1')
    assert fake_jirate.transitions(issue) == fake_transitions['transitions']

    fake_jirate.jira._session.reset()
    assert fake_jirate.move('TEST-1', 'done') == [issue]
    assert fake_jirate.jira._session.post_urls == {'https://domain.com/rest/api/2/issue/1000001/transitions': {'transition': {'id': '13'}}}

    fake_jirate.jira._session.reset()
    assert fake_jirate.move('TEST-1', 'done', resolution='won\'t do') == [issue]
    assert fake_jirate.jira._session.post_urls == {'https://domain.com/rest/api/2/issue/1000001/transitions': {'fields': {'resolution': {'name': 'Won\'t Do'}}, 'transition': {'id': '13'}}}

    fake_jirate.jira._session.reset()
    assert fake_jirate.move('TEST-1', 'done', resolution='done') == [issue]
    assert fake_jirate.jira._session.post_urls == {'https://domain.com/rest/api/2/issue/1000001/transitions': {'fields': {'resolution': {'name': 'Done'}}, 'transition': {'id': '13'}}}


def test_transition_bad_field():
    issue = fake_jirate.issue('TEST-1')
    assert fake_jirate.transitions(issue) == fake_transitions['transitions']

    fake_jirate.jira._session.reset()
    with pytest.raises(ValueError):
        assert fake_jirate.move('TEST-1', 'done', beastly_fido='odif_yltsaeb') == [issue]


@pytest.mark.parametrize("param,expected", [
    ('Fixed in Build', 'customfield_1234567'),
    ('fixed_in_build', 'customfield_1234567'),
    ('customfield_1234567', 'customfield_1234567')])
def test_field_to_id(param, expected):
    assert fake_jirate.field_to_id(param) == expected


@pytest.mark.parametrize("param,expected", [
    ('Fixed in Build', 'fixed_in_build'),
    ('fixed_in_build', 'fixed_in_build'),
    ('customfield_1234567', 'fixed_in_build')])
def test_field_to_alias(param, expected):
    assert fake_jirate.field_to_alias(param) == expected


@pytest.mark.parametrize("param,expected", [
    ('Fixed in Build', 'Fixed in Build'),
    ('fixed_in_build', 'Fixed in Build'),
    ('customfield_1234567', 'Fixed in Build')])
def test_field_to_human(param, expected):
    assert fake_jirate.field_to_human(param) == expected



#!/usr/bin/python3
#
# Jinja2 variable support for Jirate templates
#

import datetime

from jinja2 import Environment, BaseLoader, meta, nodes


def update_values_interactive(values, cli_values):
    """
    Grab input from the user for all values that do not already have one
    as input from the CLI.

    If a default is provided and no input is given, accept the default.

    Parameters
    ----------
    values : dict
        Variable names and defaults values from the Jinja2 template
    cli_values : dict
        Variable names and values provided from the command line

    Returns
    -------
    dict
        Variable name and value pairs
    """
    ret = {}
    for key in values:
        # If provided via CLI, move on
        if key in cli_values:
            ret[key] = cli_values[key]
            continue

        # If a default was set in the jinja2 template, start with that
        if values[key]:
            ret[key] = input(f'Value for "{key}" (default: "{values[key]}"): ')
        else:
            ret[key] = input(f'Value for "{key}": ')
        if not ret[key]:
            ret[key] = values[key]

    return ret


# In Jinja2, you can set a default globally like this:
# {% set var = "value" %}
# But, if you do that, you no longer can override the default
# when you render the template.
# You also can set a default per variable instance, but we will
# overwrite it with the first.  So, for setting a default value
# in a template, you can do:
#
# {% set var=var or "value" %}  # form 1; code at top of file
# {{var|default("value")}}      # form 2; inline var w/ default
def __assign_default(node, ret):
    """
    Grab the assignment variable name and default value from an Assign node.
    In a Jinja2 document, this looks like:
        {% set version=version or "1.0" %}
    ...which results in the following in the Jinja2 tree:
        Assign(target=Name(name='abc', ctx='store'), node=Or(left=Name(name='version', ctx='load'), right=Const(value='1.0'))),

    Parameters
    ----------
    node : jinja2.nodes.Or
        Current node in Jinja2 abstract syntax tree
    ret : dict
        Return variable names and default values (linear)

    Returns
    -------
    bool
        True if a value was assigned, else False
    """
    if not isinstance(node.node, nodes.Or):
        return False
    if not isinstance(node.node.left, nodes.Name):
        return False
    if not isinstance(node.node.right, nodes.Const):
        return False

    # Ignore template variables that start with underscore
    if node.target.name.startswith('_'):
        return False

    ret[node.target.name] = node.node.right.value
    return True


def __filter_default(node, ret):
    """
    Grab the assignment variable name and default value from an Assign node.
    In a Jinja2 document, this looks like:
        {{version|default('1.0')}}
    ...which results in the following in the Jinja2 tree:
        Filter(node=Name(name='version', ctx='load'), name='default', args=[Const(value='1.0')

    Parameters
    ----------
    node : jinja2.nodes.Filter
        Current node in Jinja2 abstract syntax tree
    ret : dict
        Return variable names and default values (linear)

    Returns
    -------
    bool
        True if a value was assigned, else False
    """
    if not isinstance(node, nodes.Filter):
        return False
    if not isinstance(node.node, nodes.Name):
        return False
    if node.name != 'default':
        return False
    if len(node.args) > 1:
        return False
    if not isinstance(node.args[0], nodes.Const):
        return False

    # Ignore template variables that start with underscore
    if node.node.name.startswith('_'):
        return False

    ret[node.node.name] = node.args[0].value
    return True


def __assemble_from_tree(node, ret):
    """
    Assembles set of defaults from a parsed Jinja2 node (recursive)

    Parameters
    ----------
    node : jinja2.nodes.*
        Jinja2 nodes class object (tree)
    ret : dict of return variable names and default values (linear)

    Returns
    -------
    dict
        variable names and default values from document tree
    """
    if isinstance(node, nodes.Assign):
        if __assign_default(node, ret):
            # Successfully found a mutable variable in form 1
            return ret
    elif isinstance(node, nodes.Output):
        for item in node.nodes:
            __assemble_from_tree(item, ret)
    elif isinstance(node, nodes.Filter):
        if __filter_default(node, ret):
            # Successfully found a mutable variable in form 2
            return ret
    elif isinstance(node, nodes.Template):
        for item in node.body:
            __assemble_from_tree(item, ret)

    return ret


def assemble_from_tree(tree):
    """
    Assembles set of defaults from a parsed Jinja2 document abstract
    syntax tree.

    Parameters
    ----------
    tree : jinja2.nodes.Template
        Jinja2 tree object
    ret : dict
        Variable names and default values (linear)

    Returns
    -------
    dict
        Variable names and default values
    """
    ret = {}  # initialize
    return __assemble_from_tree(tree, ret)


def apply_values(inp, values={}, interactive=False):
    """
    Apply Jinja2 variable defaults (lowest priority), interactive
    inputs (medium priority) or values provided on the command line
    (highest priority) and to a Jinja2 template and return the
    rendered text.

    Note that Jinja2 variable substitution uses glyphs that are native
    to JIRA so we use a modified set ({{ -> {@, }} -> @})

    Parameters
    ----------
    inp : str
        Raw Jinja2 template as text
    values : dict
        Values provided from the command line, if any
    interactive : bool
        If false, do NOT prompt for user input; True means to prompt
        the user for input.

    Returns
    -------
    str
        Rendered text template.

    Raises
    ------
    ValueError
        No value for a variable or no variable matching input in
        template.
    """
    env = Environment(loader=BaseLoader,
                      trim_blocks=True, lstrip_blocks=True,
                      variable_start_string='{@',
                      variable_end_string='@}')
    env.globals['datetime'] = datetime.datetime
    jinja_template = env.from_string(inp)
    ast = env.parse(inp)

    # Pass 1: store all values w/ defaults
    template_values = assemble_from_tree(ast)

    # pass 2: store unassigned variables too
    unset_keys = meta.find_undeclared_variables(ast)
    for key in unset_keys:
        # Ignore all variables that start with underscore
        if key.startswith('_'):
            continue
        if key not in template_values:
            template_values[key] = ''

    if interactive:
        values = update_values_interactive(template_values, values)

    extra = []
    for key in values:
        if key not in template_values:
            extra.append(key)
        template_values[key] = values[key]
    if extra:
        raise ValueError(f'Unknown variable(s) {extra}')

    missing = []
    for key in template_values:
        if not template_values[key]:
            missing.append(key)
    if missing:
        raise ValueError(f'Missing value(s) for {missing}')

    return jinja_template.render(template_values)



#!/usr/bin/python3

from jirate.decor import truncate, parse_params, comma_separated


def test_parse_simple():
    assert parse_params('arg') == ['arg']


def test_parse_commas():
    assert parse_params('arg,barg') == ['arg', 'barg']
    assert parse_params('"Comma, Test, ", 2') == ['Comma, Test,', "2"]


def test_parse_spaces():
    assert parse_params('"Space Test"') == ['Space Test']
    assert parse_params('"Space Test",2') == ['Space Test', '2']
    assert parse_params('"Space Test","2"') == ['Space Test', "2"]


def test_parse_quote_eating():
    assert parse_params('"Space Test":10') == ['Space Test:10']


def test_truncate_identity():
    assert truncate('abc', 0) == 'abc'
    assert truncate('abcdef', 10) == 'abcdef'


def test_truncate_exceed():
    assert truncate('abc', 3) == 'abc'
    assert truncate('abcd', 3) == 'ab…'


def test_truncate_negative():
    assert truncate('abc', -3) == 'abc'
    assert truncate('abcd', -3) == '…cd'


def test_comma_separated():
    str_val = 'one, two, "three, and four"'
    list_val = ['one', 'two', 'three, and four']

    assert comma_separated(list_val) == str_val


def test_parse_and_unparse():
    str_val = 'one, two, "three, and four"'
    list_val = ['one', 'two', 'three, and four']

    assert parse_params(comma_separated(list_val)) == list_val
    assert comma_separated(parse_params(str_val)) == str_val



#!/usr/bin/python3

import re


_val_to_py = {
    'false': False,
    'true': True,
    '<null>': None
}


def val_to_py(val):
    if val in _val_to_py:
        return _val_to_py[val]
    try:
        ret = int(val)
        return ret
    except ValueError:
        pass
    try:
        ret = float(val)
        return ret
    except ValueError:
        pass
    return val


# Jira Sprint Rendering
# TODO: Allow custom sprint output
def sprint_content_to_py(sprint_info):
    ret = []
    if isinstance(sprint_info, str):
        sprint_info = [sprint_info]

    for sprint in sprint_info:
        match = re.match(r'com.atlassian.greenhopper.service.sprint.Sprint@([a-z0-9]+)\[([^\]]+)\]', sprint)
        if not match:
            raise ValueError(f'Could not parse sprint: {sprint}')

        item = {'_hash': str(match.group(1))}
        info_bits = match.group(2)
        for bit in info_bits.split(','):
            try:
                key, value = bit.split('=')
                item[key] = val_to_py(value)
            except ValueError:
                continue
        ret.append(item)
    return ret


def sprint_field(data, fields, as_object=False):
    if as_object:
        return data
    sprints = sprint_content_to_py(data)
    active_sprints = []
    for sprint in sprints:
        if sprint['state'] == 'ACTIVE':
            active_sprints.append(f"{sprint['name']} (ID: {sprint['id']})")

    if active_sprints:
        # All active sprints
        return ', '.join(active_sprints)

    # Last closed sprint
    sprint = sprints[-1]
    return f"{sprint['name']} (ID: {sprint['id']})"


def no_display(data, fields, as_object=False):
    if as_object:
        return data
    return None


# Used by jira_fields
custom_field_renderers = {
    'com.atlassian.jira.plugins.jira-development-integration-plugin:devsummary': no_display,
    'com.onresolve.jira.groovy.groovyrunner:scripted-field': no_display,
    'com.pyxis.greenhopper.jira:gh-epic-color': no_display,
    'com.pyxis.greenhopper.jira:gh-global-rank': no_display,
    'com.pyxis.greenhopper.jira:gh-lexo-rank': no_display,
    'com.pyxis.greenhopper.jira:gh-sprint': sprint_field
}



#!/usr/bin/python3
import copy
import os
import time

import pytest  # NOQA

from jirate.rqcache import RequestCache
import jirate.rqcache


class TestSession(object):
    def request(self, method, url, **kwargs):
        return {'value': os.urandom(16), 'resp': 200}

    def get(self, url, **kwargs):
        return self.request('GET', url, **kwargs)

    def put(self, url, **kwargs):
        return self.request('PUT', url, **kwargs)
    """ Uncomment when needed; disabled for coverage
    def post(self, url, **kwargs):
        return self.request('POST', url, **kwargs)
    """


# Overwrite ResilientSession import
jirate.rqcache.ResilientSession = TestSession


def test_rqcache_nomatch():
    session = TestSession()
    cache = RequestCache(session, filename=None, expire=10)  # NOQA
    # url1 is not matched by our paths so it should always
    # be a cache miss
    ret1 = session.get('url1')
    ret2 = session.get('url1')
    assert ret1 != ret2

    session.put('url2')


def test_rqcache_match():
    session = TestSession()
    cache = RequestCache(session, filename=None, expire=10)  # NOQA
    # /field is a match, so we should get the cached data on
    # the second pull
    ret1 = session.get('https://whatever/rest/api/2/field', params={'data': '1'})
    ret2 = session.get('https://whatever/rest/api/2/field', params={'data': '2'})
    ret3 = session.get('https://whatever/rest/api/2/field', params={'data': '1'})
    ret4 = session.get('https://whatever/rest/api/2/field', params={'data': '2'})
    ret5 = session.get('https://whatever/rest/api/2/field', params={'data': '3'})

    assert ret1 != ret2
    assert ret1 == ret3
    assert ret2 == ret4
    assert ret1 != ret5
    assert ret2 != ret5

    cache.debug_dump()


def test_rqcache_match_expire():
    session = TestSession()
    cache = RequestCache(session, filename=None, expire=1)  # NOQA
    # We expire in 1 second, so second pull should have new
    # data
    ret1 = session.get('https://whatever/rest/api/2/field')
    time.sleep(1.1)
    ret2 = session.get('https://whatever/rest/api/2/field')

    assert ret1 != ret2


def test_rqcache_purge_expired():
    session = TestSession()
    cache = RequestCache(session, filename=None, expire=1)
    # Here, our baseline has no requests
    baseline = copy.deepcopy(cache.cached_reqs)
    session.get('https://whatever/rest/api/2/field')
    # Now our cache has one request, which expires in 1 second
    # Flushing the cache should not matter
    cache.flush()
    assert cache.cached_reqs != baseline

    # Wait for our req to expire
    time.sleep(1.1)
    cache.flush()
    # Now, we should be back to baseline
    assert cache.cached_reqs == baseline


def test_rqcache_load_none(tmp_path):
    session = TestSession()
    cache = RequestCache(session, filename=None, expire=1)
    assert not cache.load(os.path.join(tmp_path, 'cache_test'))
    assert not cache.load()


def test_rqcache_load_bad(tmp_path):
    session = TestSession()
    filename = os.path.join(tmp_path, 'cache_test')
    with open(filename, 'w') as fp:
        fp.write('hello, world!')
    cache = RequestCache(session, filename, expire=1)
    # Tests using our initial filename
    assert not cache.load()

    # Bad magic # Note: doing this to your cache then
    # reading it, your object's behavior is undefined
    del cache.cached_reqs['magic']
    assert cache.save() == None  # pickle.dump returns None
    # This will wipe the file
    cache.load()

    # We unlink the file if it's bad data
    with pytest.raises(FileNotFoundError):
        os.unlink(filename)

    # No save
    cache._cache_file = None
    assert cache.save() == None


def test_rqcache_persist(tmp_path):
    session = TestSession()
    filename = os.path.join(tmp_path, 'cache_test')
    cache = RequestCache(session, filename=filename)  # use default
    ret1 = session.get('https://whatever/rest/api/2/field')
    cache.save()

    # Load our cache from disk
    session2 = TestSession()  # NOQA
    cache2 = RequestCache(session, filename=filename, expire=1)  # NOQA
    ret2 = session.get('https://whatever/rest/api/2/field')

    # Request should match
    assert ret1 == ret2


def test_rqcache_persist_expire(tmp_path):
    session = TestSession()
    filename = os.path.join(tmp_path, 'cache_test')
    cache = RequestCache(session, filename=filename, expire=1)
    ret1 = session.get('https://whatever/rest/api/2/field')
    cache.save()
    time.sleep(1.1)

    # Load our cache from disk
    session2 = TestSession()  # NOQA
    cache2 = RequestCache(session, filename=filename, expire=1)  # NOQA
    ret2 = session.get('https://whatever/rest/api/2/field')

    # Request should not match
    assert ret1 != ret2


def test_user_break(tmp_path):
    session = TestSession()
    filename = os.path.join(tmp_path, 'cache_test')
    cache = RequestCache(session, filename=filename, expire=1)
    url = 'https://whatever/rest/api/2/field'
    cache.user_breaks = {'GET': [url]}
    with pytest.raises(Exception):
        session.get(url)




#!/usr/bin/python3

import json
import yaml
import os


class ParseError(Exception):
    pass


def _str_presenter(dumper, data):
    """
    Makes PyYAML print multiline strings in a sensible way
    """
    data = data.rstrip()
    lines = data.splitlines()
    if len(lines) > 1:
        data = '\n'.join([line.rstrip() for line in lines])
        return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
    return dumper.represent_scalar('tag:yaml.org,2002:str', data)


def yaml_dump(info):
    yaml.representer.SafeRepresenter.add_representer(str, _str_presenter)
    return yaml.safe_dump(info, allow_unicode=True)


# Allow YAML or JSON
def _auto_parse(config_data):
    try:
        return json.loads(config_data)
    except json.decoder.JSONDecodeError:
        pass

    try:
        return yaml.safe_load(config_data)
    except (yaml.scanner.ScannerError, yaml.reader.ReaderError):
        pass

    raise ParseError('Could not parse configuration file')


def get_config(filename=None):
    config_file = None
    if not filename:
        for filename in ['~/.jirate.json', '~/.jirate.yaml', '~/.trolly.json']:
            try:
                config_file = open(os.path.expanduser(filename))
                break
            except FileNotFoundError:
                continue
    else:
        try:
            config_file = open(os.path.expanduser(filename))
        except FileNotFoundError:
            raise

    if not config_file:
        raise FileNotFoundError('No valid configuration file found')

    config_data = config_file.read()
    config_file.close()
    config = _auto_parse(config_data)

    return config



#!/usr/bin/python3

import copy
import os
import re
import sys
import types

from toolchest.strutil import list_or_splitstr

from jira import JIRA, JIRAError
from jira.utils import json_loads as _json_loads
from jira.resources import Issue, User

from jirate.decor import nym
from jirate.jira_input import transmogrify_input


# lhh - seems python 3.12.4 doesn't let us simply replace
# this, so do it the hard way and detect being run under
# pytest
_test_ = False
if "pytest" in sys.modules:
    _test_ = True


def json_loads(val):
    if _test_:
        return val
    return _json_loads(val)


def _user_fix(obj, *args):
    # JIRA.user() doesn't search by key
    try:
        ret = JIRA.user(obj, *args)
        return ret
    except JIRAError:
        # There's no check by key; try it
        pass

    return _user_by_key(obj, *args)


def _user_by_key(obj, *args):
    key = args[0]
    data = {}
    if len(args) > 1:
        data = {'expand': args[1]}
    user = User(obj._options, obj._session, _query_param='key')
    user.find(key, params=data)
    return user


def _resolve_field(obj, field_name):
    return obj._jirate.field(obj, field_name)


# Intelligent field update for Issue
def _update_field(issue, field_name_human, value_human, operation='set', fields=None):
    if not fields:
        # TODO use native python-jira issue.fields instead of raw json
        # (Except operations are not captured, which we need)
        fields = issue._jirate.fields(issue.key)
    if not isinstance(value_human, list):
        value_human = str(value_human)

    # TODO multi-field sets?
    field_args = {field_name_human: value_human}
    output_args = transmogrify_input(fields, **field_args)

    if not output_args:
        raise AttributeError(f'No field like \'{field_name_human}\' in {issue.key}')

    # Set up for the rest
    field_ids = [key for key in output_args.keys()]
    field_id = field_ids[0]
    field = fields[field_id]
    send_val = output_args[field_id]

    ops = field['operations']
    if operation not in ops:
        raise ValueError(f'Cannot perform \'{operation}\' on \'{field_name_human}\' of {issue.key}; try: {ops}')

    # Add and remove use a different format than 'set'.
    # There's also 'modify', but ... that one's even more complicated.
    if operation in ['add', 'remove']:
        update_args = {field_id: [{operation: val} for val in send_val]}
    else:
        update_args = {field_id: [{operation: send_val}]}

    return issue.update(**update_args)


def _resolve_field_setup(jirate_obj, issue_obj):
    # Do NOT trample future python JIRA objects' field function.
    if hasattr(issue_obj, 'field'):
        raise Exception('API BREAK: \'field\' is now part of jira.resources.Issue. Please file a bug against Jirate!')
    if hasattr(issue_obj, 'update_field'):
        raise Exception('API BREAK: \'update_field\' is now part of jira.resources.Issue. Please file a bug against Jirate!')
    issue_obj._jirate = jirate_obj
    issue_obj.field = types.MethodType(_resolve_field, issue_obj)
    issue_obj.update_field = types.MethodType(_update_field, issue_obj)


def _check_fields(issue, name):
    if name in issue.raw['fields']:
        return name
    if nym(name) in issue.raw['fields']:
        return nym(name)
    for field in issue.raw['fields']:
        if nym(name) == nym(field):
            return field
    return None


class Jirate(object):
    """High-level wrapper for python-jira"""

    def __init__(self, jira):
        self.jira = jira
        self._field_to_id = None
        self._field_to_alias = None
        self._field_to_human = None
        jira.user = types.MethodType(_user_fix, jira)
        jira.user_by_key = types.MethodType(_user_by_key, jira)

    def _issue_key(self, alias):
        if isinstance(alias, str):
            return alias.upper()
        elif isinstance(alias, Issue):
            return alias.key
        elif isinstance(alias, int):
            return str(alias)
        raise ValueError(f'Unhandled type for {alias}')

    @property
    def user(self):
        return self.jira.myself()

    def attach(self, issue_alias, url, description):
        """Attach an external URL to an issue

        Parameters:
          issue_alias: int or string, could be JIRA Issue ID or key
          url: URL to attach (string)
          description: Description of link (string)
        """
        item = {'url': url, 'title': description}
        issue = self.issue(issue_alias)
        return self.jira.add_simple_link(issue, item)

    def search_users(self, username):
        """Wrapper for searching JIRA's user database

        Parameters:
          username: Partial username or email (string)

        Returns:
          list of jira.resources.User
        """
        # Search userlist for a username.  This is provided like this
        # so we can expand functionality later. Max is 50 by default;
        # we'll start with that
        users = self.jira.search_users(username)
        return users

    def field_to_id(self, alias_or_human):
        """
        Maps alias and human-readable field names to internal field identifiers.

        Parameters:
          alias_or_human: human-readable field name like "Story Points" or field alias like "story_points" (string)

        Returns:
          field ID like "customfield_12310243" (string)
        """
        if self._field_to_id is None:
            self._field_map_init()
        if alias_or_human in self._field_to_id:
            return self._field_to_id[alias_or_human]
        return None

    def field_to_alias(self, id_or_human):
        """
        Maps internal field identifiers and full human-readable names to field aliases

        Parameters:
          id_or_human: field ID like "customfield_12310243" or human-readable name like "Story Points" (string)

        Returns:
          field alias like "story_points" (string)
        """
        if self._field_to_alias is None:
            self._field_map_init()
        if id_or_human in self._field_to_alias:
            return self._field_to_alias[id_or_human]
        return None

    def field_to_human(self, id_or_alias):
        """
        Maps internal field identifiers and field aliases to human-readable field names

        Parameters:
          id_or_alias: field ID like "customfield_12310243" or alias like "story_points" (string)

        Returns:
          human-readable field name like "Story Points" (string)
        """
        if self._field_to_human is None:
            self._field_map_init()
        if id_or_alias in self._field_to_human:
            return self._field_to_human[id_or_alias]
        return None

    def _builtin_map_init(self, jira_val, human_val):
        self._field_to_id[jira_val] = jira_val
        self._field_to_id[human_val] = jira_val
        self._field_to_alias[jira_val] = jira_val
        self._field_to_alias[human_val] = jira_val
        self._field_to_human[jira_val] = human_val
        self._field_to_human[human_val] = human_val

    def _field_map_init(self):
        # For inscrutable reasons Jira returns all possible fields via the /field API...
        # ...all but one: the "parent" field. We hardcode the translation so higher-level
        # code doesn't need to deal with that.
        self._field_to_id = {}
        self._field_to_alias = {}
        self._field_to_human = {}
        self._builtin_map_init('parent', 'Parent')
        self._builtin_map_init('fixVersions', 'fixversions')
        self._builtin_map_init('lastViewed', 'lastviewed')

        fields = self.jira.fields()
        for field in fields:
            field_id = field['id']
            name = field['name']
            alias = nym(name)
            # append underscores for collisions
            # XXX hopefully this is extremely rare
            while alias in self._field_to_id:
                alias = alias + '_'
            # Everything maps to everything. _field_to_id can return a field ID when fed
            # either the human name, the alias, or even the ID itself.
            for val in (field_id, name, alias):
                self._field_to_id[val] = field_id
                self._field_to_human[val] = name
                self._field_to_alias[val] = alias
            for clause_name in field['clauseNames']:
                if (re.match('^cf\\[[0-9]+\\]$', clause_name) or clause_name in self._field_to_id):
                    # Skip nonsense and duplicate alternative names
                    continue
                self._field_to_id[clause_name] = field_id
                self._field_to_human[clause_name] = name
                self._field_to_alias[clause_name] = alias

    def search_issues(self, search_query):
        """Run a JQL search and assemble the results into one list

        Parameters:
          search_query: JQL query line (string)

        Returns:
          list of jira.resources.Issue
        """
        index = 0
        chunk_len = 50      # So we can detect end
        ret = []
        while True:
            issues = self.jira.search_issues(search_query, startAt=index, maxResults=chunk_len)
            if not len(issues):
                break
            ret.extend(issues)
            index = index + len(issues)
            if len(issues) < chunk_len:
                break
        for issue in ret:
            _resolve_field_setup(self, issue)
        return ret

    def _field(self, issue, field_name):
        """Reconcile a field in an issue with custom field defs
        on the jira server or an issue's fields. Does not retrieve
        the issue from the JIRA server.

        Parameters:
          issue_alias: int or string, could be JIRA Issue ID or key
          field_name: human readable field name

        Returns:
          field value, or none if not found

        Raises:
          AttributeError if the field does not exist.
        """
        # Don't return customfield if there's a direct match
        # obvious match
        if field_name in issue.raw['fields']:
            return field_name
        fname = self.field_to_id(field_name)
        if fname in issue.raw['fields']:
            return fname
        fname = _check_fields(issue, field_name)
        if fname in issue.raw['fields']:
            return fname
        raise AttributeError(str(issue) + f' has no field like {field_name}')

    def field(self, issue_alias, field_name):
        """Reconcile a field in an issue with custom field defs
        on the jira server. Retrieves the issue from the server if
        needed.

        Parameters:
          issue_alias: int or string, could be JIRA Issue ID or key
          field_name: human readable field name

        Returns:
          field value, or none if not found

        Raises:
          AttributeError if the field does not exist.
        """
        issue = self.issue(issue_alias)
        fname = self._field(issue, field_name)
        return issue.raw['fields'][fname]

    def fields(self, issue_alias):
        """Determine the fields available for an issue

        Parameters:
          issue_alias: int or string, could be JIRA Issue ID or key

        Returns:
          dict of fields
        """
        issue = self.issue(issue_alias)
        # XXX HERE THERE BE DRAGONS
        # NOT IMPLEMENTED UPSTREAM
        url = os.path.join(issue.raw['self'], 'editmeta')
        field_blob = json_loads(self.jira._session.get(url))
        return field_blob['fields']

    def get_user(self, username):
        """Determine JIRA's normalized username for someone

        Parameters:
          username: email, display name or username

        Returns:
          username (string)
        """
        if not username or username.lower() == 'none':
            return None
        if username == 'me':
            return self.user['name']
        users = self.jira.search_users(username)
        if len(users) > 1:
            raise ValueError(f'Multiple matching users for \'{username}\'')
        elif not users:
            raise ValueError(f'No matching users for \'{username}\'')

        return users[0].name

    def api_call(self, uri, raw=False):
        url = self.jira._get_url(uri)
        ret = self.jira._session.get(url)
        if raw:
            return ret.text
        return json_loads(ret)

    def assign(self, issue_aliases, users=None):
        """Assign a set of issues to a user

        NOTE: Originally, the additional users were going to be set as
        watchers, which is another API call. We did this with Trello.
        Do we want to do that for JIRA, or should we have separate
        watch/unwatch?

        Parameters:
          issue_aliases: list of issue keys or IDs (list of strings)
          users: list of users (list of strings)

        Returns:
          list of issues successfully assigned to the user(s)
        """
        if isinstance(users, str):
            users = [users]
        issues = self.issues(issue_aliases)

        for issue in issues:
            user_ids = []

            # first is assignee
            if users:
                for user in users:
                    uid = self.get_user(user)
                    if uid not in user_ids:
                        user_ids.append(uid)
            else:
                # Just me
                user = self.user['name']
                user_ids = [user]
                # user_ids = [user['id']]

            if not len(users):
                return

            user = user_ids.pop(0)
            issue.update(assignee=user)

    def sprint_info(self, project_key, states=['active', 'future']):
        """Retrieve all sprints and boards for a project.

        Parameters:
          project_key: Key of project to check
          states: Array or string (comma separated) of sprint states

        Returns:
          dict with boards and sprints
        """
        if isinstance(states, list):
            states = ','.join(states)

        ret = {'boards': {}, 'sprints': {}}
        # XXX fixme: paginated APIs
        boards = []
        _start = 0
        _max = 50
        while True:
            new_boards = self.jira.boards(projectKeyOrID=project_key, startAt=_start, maxResults=_max)
            if not new_boards:
                break
            _start = _start + _max
            boards.extend(new_boards)

        sprints = []
        for board in boards:
            if board.type != 'scrum':
                continue
            _start = 0
            while True:
                new_sprints = self.jira.sprints(board.id, startAt=_start, maxResults=_max, state=states)
                if not new_sprints:
                    break
                _start = _start + _max
                sprints.extend(new_sprints)

            if board.name in ret['boards']:
                old = ret['boards'][board.name]
                if old.id != board.id:
                    print(f'Warning: Duplicate Board: {old.name} (IDs: {old.id} {board.id})')
            ret['boards'][board.name] = board

        for sprint in sprints:
            if sprint.name in ret['sprints']:
                old = ret['sprints'][sprint.name]
                if old.id != sprint.id:
                    print(f'Warning: Duplicate Sprint: {old.name} (IDs: {old.id} {sprint.id})')
            ret['sprints'][sprint.name] = sprint
        return ret

    def get_comment(self, issue_alias, comment_id):
        """Retrieve raw comment for an issue

        Parameters:
          issue_alias: issue key or ID (string)
          comment_id: ID of comment to retrieve

        Returns:
          jira.resources.Comment
        """
        issue = self.issue(issue_alias)
        return self.jira.comment(issue.raw['key'], comment_id)

    def comment(self, issues, text, visibility=None):
        """Attach a new comment to a set of issues

        Parameters:
          issues: issue key or ID (string)
          text: Text to attach as comment

        Returns:
          [jira.resources.Issue] for updated issues
        """
        issue_list = list_or_splitstr(issues)

        comment_data = {'body': text}
        if visibility:
            if isinstance(visibility, str):
                comment_data['visibility'] = {}
                comment_data['visibility']['type'] = 'group'
                comment_data['visibility']['value'] = visibility
            elif isinstance(visibility, dict):
                comment_data['visibility'] = visibility

        ret = []
        for alias in issue_list:
            issue = self.issue(alias)
            # Use simple comment mode to add a comment
            url = os.path.join(issue.raw['self'], 'comment')
            if self.jira._session.post(url, data=comment_data):
                ret.append(issue)
        return ret

    def close(self, issues, **args):
        """Close an issue

        XXX this might not be a usable API and/or may be a higher
        level transition than we want. Note that 'Closed' may also
        not be correct for a given project, since all transitions
        are per-project. We could do something intelligent with
        move, or delete this API altogether.

        Parameters:
          issues: list of issue keys or IDs (list of string)
          args: dictionary of fields to set on transition
                (resolution only really known right now)

        Returns:
          requests.Response
        """
        return self.move(issues, 'Closed', **args)

    def create(self, field_definitions=None, **args):
        """Create a new issue using key/value pairs

        Parameters:
          field_definitions: List of creation definitions for the
                             issue type.
          **args: Dictionary of key/value pairs (dict)

        Returns:
          jira.resources.Issue
        """
        # Make sure we create in correct project if we're creating
        # a subtask. Also resolve parent issue.
        if 'parent' in args and isinstance(args['parent'], str):
            parent_issue = self.issue(args['parent'])
            project = parent_issue.raw['fields']['project']['key']
            args['parent'] = parent_issue.key
            args['project'] = project

        # Transmogrify other fields
        new_args = transmogrify_input(field_definitions, **args)
        return self.jira.create_issue(**new_args)

    def update_issue(self, issue_alias, field_definitions=None, **kwargs):
        """Update an issue using key/value pairs

        Parameters:
          **args: Dictionary of key/value pairs (dict)

        Returns:
          jira.resources.Issue (?)
        """
        issue = self.issue(issue_alias)
        if not issue:
            return None

        args = {}
        for field in kwargs:
            args[self.field_to_id(field)] = kwargs[field]
        return issue.update(**kwargs)

    def update(self, issue_list, **kwargs):
        """Update a set of issues using key/value pairs

        Parameters:
          **kwargs: Dictionary of key/value pairs (dict)

        Returns:
          [jira.resources.Issue] of updated issues
        """
        ret = []
        args = {}

        for field in kwargs:
            args[self.field_to_id(field)] = kwargs[field]
        issues = list_or_splitstr(issue_list)
        for issue_alias in issues:
            issue = self.issue(issue_alias)
            if not issue:
                continue
            issue.update(**args)
            ret.append(issue)
        return ret

    def issue(self, issue_alias, verbose=False):
        """Retrieve an issue from JIRA

        XXX Cleanup vs. JiraProject

        Parameters:
          issue_alias: key or issue ID (string)

        Returns:
          jira.resources.Issue
        """
        if isinstance(issue_alias, Issue):
            return issue_alias
        issue_aliases = [issue_alias]
        if isinstance(issue_alias, int):
            issue_alias = str(issue_alias)
        if issue_alias.upper() != issue_alias:
            issue_aliases.append(issue_alias.upper())
        for alias in issue_aliases:
            try:
                issue = self.jira.issue(alias)
                if not issue:
                    continue
                _resolve_field_setup(self, issue)
                return issue
            except JIRAError:
                pass
        return None

    def votes(self, issue_alias):
        if isinstance(issue_alias, Issue):
            ret = self.jira.votes(issue_alias.key)
            if ret:
                issue_alias.raw['fields']['votes'] = ret.raw
        else:
            ret = self.jira.votes(issue_alias)
        return ret

    def watchers(self, issue_alias):
        if isinstance(issue_alias, Issue):
            ret = self.jira.watchers(issue_alias.key)
            if ret:
                issue_alias.raw['fields']['watches'] = ret.raw
        else:
            ret = self.jira.votes(issue_alias)
        return ret

    def eausm_issue_votes(self, issue_alias):
        """Retrieve EAUSM (Easy Agile Planning Poker) votes
        for an issue

        Parameters:
          issue_alias: key, issue ID (string), or Issue object

        Returns:
          raw json data of EAUSM information
        """
        issue = self.issue(issue_alias)
        if not issue:
            return None
        if 'eausm' in issue.raw['fields']:
            return issue.raw['fields']['eausm']

        # Check for the EZ Agile Planning Poker ext on the server
        EAUSM_url = self.jira.server_url + \
            f"/rest/eausm/latest/planningPoker/{issue.id}"
        try:
            ret = self.jira._session.get(EAUSM_url)
            EAUSM_json = json_loads(ret)
            issue.raw['fields']['eausm'] = EAUSM_json
        except JIRAError:
            return False
        return issue.raw['fields']['eausm']

    def eausm_vote_issue(self, issue_alias, vote):
        """Set EAUSM (Easy Agile Planning Poker) votes
        for an issue

        Parameters:
          issue_alias: key, issue ID (string), or Issue object
          votes: string or integer for story point voting

        Returns:
          None: no such issue or voting disabled
          False: Failed to set
          True: OK
        """
        issue = self.issue(issue_alias)
        if not issue:
            return None
        vote = int(vote)

        EAUSM_url = f"{self.jira.server_url}/rest/eausm/latest/planningPoker/vote"
        payload = {"issueId": issue.id, "vote": vote}
        ret = self.jira._session.put(EAUSM_url, data=payload)
        if not ret:
            return False
        return True

    def issues(self, issue_list, verbose=False):
        """Retrieve one or more issues from JIRA

        Parameters:
          issue_list: string of keys or list of keys (strings)

        Returns:
          list of jira.resources.Issue or None
        """
        if not issue_list:
            return []
        if isinstance(issue_list, Issue):
            return issue_list
        issues = list_or_splitstr(issue_list)
        search_issues = []
        issue_objs = []
        for issue in issues:
            if isinstance(issue, Issue):
                issue_objs.append(issue)
            else:
                search_issues.append(self._issue_key(issue))

        if len(search_issues) == 1:
            # If we only need one issue, avoid the risk of the extra call to grab fields
            ret = [self.issue(search_issues[0])]
        else:
            # This is one API call instead of N (two if fields have not been cached)
            ret = self.search_issues('key in (' + ', '.join(search_issues) + ')')
        ret.extend(issue_objs)
        return ret or None

    def transitions(self, issue):
        """Retrieve possible next-state transitions for an issue

        Parameters:
          issue_alias: key or issue ID (string)

        Returns:
          list of transitions w/ metadata
        """
        if isinstance(issue, str):
            issue = self.issue(issue)
        if not issue:
            return None
        url = os.path.join(issue.raw['self'], 'transitions?expand=transitions.fields')
        transitions = json_loads(self.jira._session.get(url))
        if transitions:
            return transitions['transitions']

    def _find_transition(self, issue, status):
        transitions = self.transitions(issue)
        for transition in transitions:
            trans_state = transition['to']['name']
            trans_state_id = transition['to']['id']
            if trans_state == status or nym(trans_state) == status or str(status) == str(trans_state_id):
                return transition
        return None

    def move(self, issue_list, status, **args):
        """Execute a transition to move a set of issues to the desired status

        Jira doesn't have a status you can update; you have to retrieve possible
        transitions and satisfy those requirements. Each issue has its own transition map
        according to the issue type within a given project

        Parameters:
          issue_aliases: list keys or issue IDs (list of string)
          status: Desired status
          **args: field=value pairs to set on transition

        Returns:
          list of successfully moved issues (list of string)
        """
        issue_aliases = list_or_splitstr(issue_list)
        issues = self.issues(issue_aliases)

        if len(issues) < len(issue_aliases):
            fails = list(set(issue_aliases) - set([issue.key for issue in issues]))
            raise ValueError('No such issue(s): ' + str(fails))

        moved = []
        for issue in issues:
            if not issue:
                continue
            transition = self._find_transition(issue, status)
            if not transition:
                continue
            # API cleanup:
            # If it's already in the target status, don't apply the move
            if transition['to']['name'] == issue.fields.status.name:
                continue

            data = {'transition': {'id': transition['id']}}
            if args and 'fields' in transition:
                new_args = transmogrify_input(transition['fields'], **args)
                data['fields'] = new_args
                if new_args == {}:
                    oops = [args.keys()]
                    raise ValueError(f'field(s) not allowed in transition: {oops}')

            # POST /rest/api/2/issue/{issueIdOrKey}/transitions
            url = os.path.join(issue.raw['self'], 'transitions')
            self.jira._session.post(url, data=data)
            moved.append(issue)
        return moved

    def link_types(self):
        """Wrapper for jira.issue_link_types()"""
        return self.jira.issue_link_types()

    def link(self, left_alias, right_alias, link_text):
        """Link two isues together using the noted link type

        Parameters:
          left_alias: left key or issue IDs (string)
          right_alias: right key or issue IDs (string)
          link_text: Desired link type

        Returns:
          ???
        """
        left = self._issue_key(left_alias)
        right = self._issue_key(right_alias)
        return self.jira.create_issue_link(link_text, left, right)

    def remote_links(self, issue_alias):
        """Obtain all remote links (URLs) attached to an issue

        Parameters:
          issue_alias: key or issue IDs (string)

        Returns:
          list of jira.resources.RemoteLink
        """
        issue = self._issue_key(issue_alias)
        return self.jira.remote_links(issue)

    def unlink(self, left_alias, right_alias):
        """Break all links between to issues

        Parameters:
          left_alias: left key or issue IDs (string)
          right_alias: right key or issue IDs (string)

        Returns:
          count of links removed
        """
        left = self.issue(left_alias)
        right = self.issue(right_alias)

        if not right:
            extlink = self.jira.remote_link(left.raw['key'], str(right_alias))
            self.jira._session.delete(extlink.raw['self'])
            return 1

        info = left.raw['fields']

        if 'issuelinks' not in info or not info['issuelinks']:
            return 0

        right_name = right.raw['key']
        count = 0
        for link in info['issuelinks']:
            link_id = None
            if 'inwardIssue' in link and link['inwardIssue']['key'] == right_name:
                link_id = link['id']
            if 'outwardIssue' in link and link['outwardIssue']['key'] == right_name:
                link_id = link['id']
            if link_id:
                count = count + 1
                self.jira.delete_issue_link(link_id)
        return count


class JiraProject(Jirate):
    def __init__(self, jira, project, closed_status=None, readonly=False, allow_code=False):
        super().__init__(jira)
        self._ro = readonly
        self._config = None
        self._closed_status = closed_status
        self._project = self.jira.project(project)
        self._issue_types = None
        self.custom_fields = None
        self.project_name = project
        self.allow_code = allow_code
        self.refresh()

        if self._closed_status is None:
            # guess at common closed states
            for status in ['CLOSED', 'DONE', 'RESOLVED']:
                try:
                    self.status_to_id(status)
                    self._closed_status = status
                    break
                except KeyError:
                    pass

    def _issue_key(self, alias):
        try:
            if str(int(alias)) == alias:
                alias = f'{self.project_name}-{alias}'
        except (TypeError, ValueError):
            pass
        return super()._issue_key(alias)

    def refresh(self):
        if not self._config:
            self._config = {'states': {},
                            'issue_map': {}}

        self.refresh_lists()

    def refresh_lists(self):
        # DANGER DANGER - using private stuff because upstream doesn't have it
        spath = os.path.join(self._project._resource, 'statuses')
        url = self._project._get_url(spath.format(str(self._project)))
        status_info = json_loads(self.jira._session.get(url))
        status_ids = []
        statuses = []

        for task_type in status_info:
            for status in task_type['statuses']:
                if status['id'] not in status_ids:
                    status_ids.append(status['id'])
                    statuses.append(status)

        for item in statuses:
            val = {}
            val['name'] = item['name']
            val['id'] = item['id']
            name = nym(val['name'])
            while name in self._config['states']:
                name = name + '_'
            self._config['states'][name] = val

    def delete_issue_map(self):
        self._config['issue_map'] = {}

    def unlabel_issue(self, issue_alias, label_name):
        return False

    def status_to_id(self, status):
        status = nym(status)

        if status not in self._config['states']:
            raise KeyError('No such list: ' + status)
        if status in self._config['states']:
            return self._config['states'][status]['id']
        return status  # must be the ID

    def search_issues(self, text):
        # Override so we can index our return values
        # TODO resolve fixversions?
        if not text:
            return None
        ret = super().search_issues(text)
        self._index_issues(ret)
        return ret

    def _index_issue(self, issue):
        if issue.key not in self._config['issue_map']:
            if not hasattr(issue, '_jirate'):
                _resolve_field_setup(self, issue)
            self._config['issue_map'][issue.key] = issue

    def _index_issues(self, issues):
        if 'issue_map' not in self._config:
            self._config['issue_map'] = {}

        for issue in issues:
            self._index_issue(issue)

    def search(self, text):
        if not text:
            return None
        return self.search_issues(f'PROJECT = {self.project_name} AND statusCategory NOT IN (Done) AND (text ~ "{text}")')

    def list(self, status=None, userid=None, all_issues=False):
        if all_issues:
            project_selector = ''
        else:
            project_selector = f'PROJECT = {self.project_name} AND '

        userid = self.get_user(userid)
        if userid is None:
            assignee_selection = 'assignee is EMPTY'
        else:
            assignee_selection = f'assignee = "{userid}"'

        if status:
            issues = super().search_issues(f'{project_selector}{assignee_selection} AND STATUS = {status}')
        else:
            issues = super().search_issues(f'{project_selector}{assignee_selection} AND statusCategory NOT IN (Done)')

        self._index_issues(issues)
        return issues

    def issue(self, issue_alias, verbose=False):
        if isinstance(issue_alias, Issue):
            return issue_alias
        issue_aliases = [issue_alias]
        if issue_alias.upper() != issue_alias:
            issue_aliases.append(issue_alias.upper())
        if '-' not in issue_alias:
            issue_aliases.insert(0, self.project_name.upper() + f'-{issue_alias}')
        for alias in issue_aliases:
            if alias in self._config['issue_map']:
                return self._config['issue_map'][alias]
            try:
                issue = self.jira.issue(alias)
                if not issue:
                    continue
                self._index_issue(issue)
                return issue
            except JIRAError:
                pass
        return None

    def eausm_issue_votes(self, issue_alias):
        if 'eausm' in self._config and not self._config['eausm']:
            return None
        ret = super().eausm_issue_votes(issue_alias)
        if ret is False:
            self._config['eausm'] = False
            return None
        return ret

    def eausm_vote_issue(self, issue_alias, votes):
        if 'eausm' in self._config and not self._config['eausm']:
            return False
        return super().eausm_vote_issue(issue_alias, votes)

    def states(self):
        return copy.copy(self._config['states'])

    def new(self, name, description=None, issue_type=None, parent=None):
        # Simple New creation requires understanding what the issuetypes are,
        # which vary on a per-project basis.  This is why "create" is separate
        # (and lower-level)
        if parent:
            if issue_type != 'Sub-task':
                raise ValueError('Specifying a parent only valid for Sub-task type')

        issuetypes = self.issue_types
        resolved_issue_type = None
        if issue_type is not None:
            for itype in issuetypes:
                if itype.name.upper() == issue_type.upper():
                    resolved_issue_type = itype.name
                    break
            if not resolved_issue_type:
                raise ValueError(f'No such issue type: {issue_type}')
        else:
            resolved_issue_type = issuetypes[0].name

        args = {}
        args['project'] = self.project_name
        args['summary'] = name
        args['issuetype'] = resolved_issue_type
        if parent is not None:
            args['parent'] = parent
        if description:
            args['description'] = description

        return self.create(**args)

    def create(self, field_definitions=None, **args):
        # override so we can index our value
        if 'project' not in args:
            args['project'] = self.project_name
        if 'issuetype' not in args:
            args['issuetype'] = 'Task'

        if not field_definitions:
            metadata = self.issue_metadata(args['issuetype'])
            field_definitions = metadata['fields']
        ret = super().create(field_definitions, **args)
        self._index_issue(ret)
        return ret

    def components(self):
        """ Return list of components assigned to this project

        Returns:
           List[component]
        """
        return self.jira.project_components(self._project)

    def add_component(self, name, description=None):
        """ Add a component to the project

        Returns:
           component
        """
        return self.jira.create_component(name, self._project, description=description)

    def remove_component(self, name):
        """ Remove a component from the project. To do this, we have to run down
        the list of components in this project to find a match then pass the ID
        to the API

        Returns:
            ???
        """
        comps = self.components()
        for comp in comps:
            if comp.name == name:
                return comp.delete()
        return 1

    def subtask(self, parent, name, description=None):
        return self.new(name, description, 'Sub-task', parent)

    @property
    def issue_types(self):
        if not self._issue_types:
            self._issue_types = self._project.issueTypes
        return self._issue_types

    @property
    def versions(self):
        return self._project.versions

    # Returns a dict that JIRA should just give us.
    def issue_metadata(self, issue_type_or_id, project_key=None):
        if not project_key:
            project_key = self.project_name
        itype = None
        for issuetype in self.issue_types:
            if issuetype.id == issue_type_or_id or nym(issuetype.name) == nym(issue_type_or_id):
                itype = issuetype
        if not itype:
            return None

        fields = []
        start = 0
        chunk_len = 50
        while True:
            new_fields = self.jira.project_issue_fields(project_key, itype.id, startAt=start, maxResults=chunk_len)
            for field in new_fields:
                fields.append(field.raw)
            if new_fields.isLast:
                break
            start = start + chunk_len

        field_dict = {val['fieldId']: val for val in fields}
        metadata = {'self': itype.self, 'name': itype.name, 'id': itype.id, 'description': itype.description, 'subtask': itype.subtask, 'iconUrl': itype.iconUrl, 'fields': field_dict}
        return metadata

    def sprint_info(self, project_key=None, states=['active', 'future']):
        if not project_key:
            project_key = self.project_name
        return super().sprint_info(project_key, states)

    def config(self):
        return copy.copy(self._config)

    def get_user_data(self, key):
        if key in ('states', 'issue_map'):
            return KeyError('Reserved configuration keyword: ' + key)
        if key in self._config:
            return copy.copy(self._config[key])
        return None

    def set_user_data(self, key, userdata):
        if key in ('states', 'issue_map'):
            return KeyError('Reserved configuration keyword: ' + key)
        self._config[key] = copy.copy(userdata)


def get_jira(jconfig):
    """Wrapper to create a python-jira connection

    Parameters:
      jconfig: dict of 3 keys: url, token, proxies (optional)

    Returns:
      JIRA
    """
    if 'url' not in jconfig:
        print('No JIRA URL specified')
        return None
    if 'token' not in jconfig:
        print('No JIRA token specified')
        return None
    if 'proxies' not in jconfig:
        jconfig['proxies'] = {"http": "", "https": ""}

    return JIRA(jconfig['url'], token_auth=jconfig['token'], proxies=jconfig['proxies'])



#!/usr/bin/python3

import os

import pytest  # NOQA

from jirate.config import get_config, ParseError, yaml_dump


def test_yaml_dump():
    inp = {"test": "abc\n\nabcdef"}
    expected = '''test: |-
  abc

  abcdef
'''
    assert yaml_dump(inp) == expected


def test_nonsense_config(tmp_path):
    filename = os.path.join(tmp_path, 'config_test')

    fp = open(filename, "w+")
    fp.write(':djsaklsajkZZ{}')
    fp.close()

    with pytest.raises(ParseError):
        get_config(filename)


def test_yaml_config(tmp_path):
    filename = os.path.join(tmp_path, 'config_test')

    fp = open(filename, "w+")
    fp.write('abc:\n')
    fp.write('  def: 1\n')
    fp.close()

    assert get_config(filename) == {'abc': {'def': 1}}


def test_json_config(tmp_path, monkeypatch):
    filename = os.path.join(tmp_path, '.jirate.json')

    fp = open(filename, "w+")
    fp.write('{"abc": {"def": 1}}\n')
    fp.close()

    assert get_config(filename) == {'abc': {'def': 1}}
    # Copied from below
    monkeypatch.setattr(os.path, 'expanduser', lambda x: tmp_path / x.replace('~/', ''))
    assert get_config() == {'abc': {'def': 1}}


# Gemini Advanced
def test_get_config_no_files_found(tmp_path, monkeypatch):
    # Test the function when no default files exist
    monkeypatch.setattr(os.path, 'expanduser', lambda x: tmp_path / x.replace('~/', ''))

    with pytest.raises(FileNotFoundError):
        get_config()

    # Added explicit filename test which should fail
    with pytest.raises(FileNotFoundError):
        get_config('nonexistent')



#!/usr/bin/python3
# Wrapper to add handlers to subparsers

import argparse


class ComplicatedArgs(object):
    def __init__(self):
        self._parser = argparse.ArgumentParser()
        self._commands = {}
        self._handlers = {}
        self._subparsers = None
        self._parsed_ns = None
        self._args = {}     # Pasted in to namespaces before calling handlers

    def parser(self):
        return self._parser

    def add_arg(self, key, value):
        self._args[key] = value

    def add_argument(self, *args, **kwargs):
        self._parser.add_argument(*args, **kwargs)

    def command(self, cmd, help=None, handler=None):
        if cmd in self._commands:
            raise ValueError('Command already specified:' + str(cmd))

        if self._subparsers is None:
            self._subparsers = self._parser.add_subparsers(help='Commands (add -h for help)', dest='command')
        subparser = self._subparsers.add_parser(cmd, help=help)
        self._commands[cmd] = subparser
        if handler is not None:
            self.register_handler(cmd, handler)
        return subparser

    def register_handler(self, cmd, handler):
        if cmd not in self._commands:
            raise ValueError('Undefined command:', str(cmd))
        self._handlers[cmd] = handler

    def delete_handler(self, cmd):
        if cmd not in self._commands:
            raise ValueError('Undefined command:', str(cmd))
        if cmd in self._handlers:
            del self._handlers[cmd]

    def parse_args(self, **kwargs):
        return self._parser.parse_args(**kwargs)

    # Call our callouts with the namespace mapping we built
    def finalize(self, ns, ret=None):
        cmd = None
        try:
            if ns.command:
                cmd = ns.command
        except AttributeError:
            pass
        for key in self._args:
            setattr(ns, key, self._args[key])
        self._parsed_ns = ns
        if cmd and cmd in self._handlers:
            return self._handlers[cmd](ns)
        # Command not found
        return ret

    # Store this so we can return our handlers' return values instead of
    # our own.
    def namespace(self):
        return self._parsed_ns


# For calling cmdline functions w/o using
# argparse
class GenericArgs(dict):
    def __getattr__(self, name):
        if (name in self):
            return dict.__getitem__(self, name)
        return None
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__



#!/usr/bin/python3
#
# AI-generated (kinda) example unit tests for ComplicatedArgs

# lhh - None of these imports were right
import argparse
import pytest

from jirate.args import ComplicatedArgs


def FunctionMockery(**kwargs):
    if 'return_value' in kwargs:
        def my_func(args):
            return kwargs['return_value']
    else:
        def my_func(args):
            return None
    return my_func


@pytest.fixture
def args():
    return ComplicatedArgs()


def test_initialization(args):
    assert isinstance(args.parser(), argparse.ArgumentParser)
    assert args._commands == {}
    assert args._handlers == {}
    assert args._subparsers is None
    assert args._parsed_ns is None
    assert args._args == {}


def test_add_argument(args):
    args.add_argument('--test', type=int)
    # Needed to add args= for each line
    parsed_args = args.parse_args(args=['--test', '10'])
    assert parsed_args.test == 10


def test_add_arg(args):
    args.add_arg('extra', 'value')
    ns = argparse.Namespace()
    args.finalize(ns)
    assert ns.extra == 'value'


def test_command_and_handler(args):
    mock_handler = FunctionMockery(return_value='test_result')
    args.command('test', help='Test command', handler=mock_handler)
    parsed_args = args.parse_args(args=['test'])
    result = args.finalize(parsed_args)
    assert result == 'test_result'


def test_second_command(args):
    handler_one = FunctionMockery(return_value='test_result')
    handler_two = FunctionMockery()
    args.command('test', help='Test command', handler=handler_one)
    args.command('test-two', help='Test command', handler=handler_two)
    parsed_args = args.parse_args(args=['test-two'])
    result = args.finalize(parsed_args)
    assert result == None


def test_command_duplicate(args):
    args.command('test')
    with pytest.raises(ValueError):
        args.command('test')


def test_register_handler_undefined_command(args):
    with pytest.raises(ValueError):
        args.register_handler('undefined', lambda x: None)


def test_delete_handler(args):
    # This test, which was AI-written, exposed bugs in delete_handler.
    args.command('test', handler=FunctionMockery(return_value=1))
    parsed_args = args.parse_args(args=['test'])
    result = args.finalize(parsed_args)
    assert result == 1
    args.delete_handler('test')
    result = args.finalize(parsed_args)
    assert result is None
    with pytest.raises(ValueError):
        args.delete_handler('HAHAHAHAHAHAH')


def test_parse_args(args):
    args.add_argument('--value', type=int)
    parsed_args = args.parse_args(args=['--value', '5'])
    assert parsed_args.value == 5


def test_finalize_no_command(args):
    ns = argparse.Namespace()
    result = args.finalize(ns, 'default')
    assert result == 'default'


def test_namespace(args):
    args.add_argument('--value', type=int)
    parsed_args = args.parse_args(args=['--value', '5'])
    args.finalize(parsed_args)
    assert args.namespace().value == 5



#!/usr/bin/python3
#
# Copy/pasted from toolchest:
#   http://github.com/release-depot/toolchest

import copy
import csv
import re
import sys
import termios

from dateutil.parser import parse
from pprint import PrettyPrinter

try:
    from rich.console import Console
    from rich.markdown import Markdown

    console = Console()
    _markdown = True
except ModuleNotFoundError:
    _markdown = False
    pass

fancy_output = False
color_shift = 16
HILIGHT = '[1m'
NORMAL = '[0m'

# xterm256 color maps for trello cards
COLORS = {'green':       2,     # NOQA
          'yellow':      11,    # NOQA
          'orange':      214,   # NOQA
          'red':         9,     # NOQA
          'purple':      93,    # NOQA
          'blue':        12,    # NOQA
          'blue-gray':   12,    # NOQA
          'turquoise':   80,    # NOQA
          'light green': 119,   # NOQA
          'pink':        205,   # NOQA
          'black':       0,     # NOQA
          'sky':         25,    # NOQA
          'lime':        10,    # NOQA
          'white':       15,    # NOQA
          'default':     7 }    # NOQA


class EscapedString(str):
    def __init__(self, val):
        self._len = len(val)
        self._text = val
        self._value = val
        self._sequence = None

    def __len__(self):
        return self._len

    def _escape(self, sequence, value=None):
        if '{_value_}' not in sequence:
            raise ValueError('Cannot escape; invalid input')
        if not value:
            value = self._text
        return sequence.replace('{_value_}', value)

    def escape(self, sequence, value=None):
        self._value = self._escape(sequence, value)

    def update(self, value):
        self._value = value

    def ljust(self, width):
        if width <= len(self):
            return str(self)
        return self + (' ' * (width - len(self)))

    def __str__(self):
        return str(self._value)

    def __repr__(self):
        return str(self._value)

    def __add__(self, other):
        if (isinstance(other, EscapedString)):
            ret = EscapedString(self._text + other._text)
            ret.update(self._value + other._value)
        else:
            ret = EscapedString(self._text + str(other))
            ret.update(self._value + str(other))
        return ret


def color_string(string, color=None, bgcolor=None):
    if fancy_output is not True:
        return string

    ret_string = ''
    fg_color = ''
    bg_color = ''
    if color and color in COLORS:
        fg_color = '[38;5;{0}m'.format(COLORS[color])
    if bgcolor and bgcolor in COLORS:
        bg_color = '[48;5;{0}m'.format(COLORS[bgcolor])

    if not fg_color and not bg_color:
        return string

    ret_string = '{0}{1}{2}[0m'.format(fg_color, bg_color, string)

    ret = EscapedString(string)
    ret.update(ret_string)

    return ret


def link_string(text, url):
    if not fancy_output:
        return None

    ret = EscapedString(text)
    ret.update(f']8;;{url}\\{text}]8;;\\')
    return ret


def issue_link_string(issue_key, baseurl=None):
    if not baseurl or not fancy_output:
        return issue_key

    return link_string(issue_key, f'{baseurl}/browse/{issue_key}')


def parse_params(arg):
    if isinstance(arg, list):
        return arg

    ret = []
    next_param = ['', '']
    stuff = csv.reader(arg)
    val = ''
    for item in stuff:
        if item == next_param:
            ret.append(val.strip())
            val = ''
            continue
        val = val + ''.join(item)
    if val:
        ret.append(val.strip())
    return ret


def comma_separated(item_list):
    out = []
    for item in item_list:
        if item and ',' in item:
            out.append(f'"{item}"')
        else:
            out.append(item)
    return ', '.join(out)


def truncate(arg, maxlen):
    if arg and maxlen:
        if maxlen > 0 and len(arg) > maxlen:
            arg = arg[:maxlen - 1] + '…'
        if maxlen < 0 and len(arg) > abs(maxlen):
            arg = '…' + arg[maxlen + 1:]
    return arg


def jira2md(jira_text):
    # Replace code/noformat blocks with triple-backticks
    return re.sub(r'({code(:java)?}|{noformat})', '```', jira_text)


def md_print(markdown_text, noformat=False):
    if _markdown and not noformat:
        fixed_text = jira2md(markdown_text)
        console.print(Markdown(fixed_text))
    else:
        print(markdown_text)


def pretty_date(date_str):
    date_obj = parse(date_str)
    if 'T' not in date_str:
        fmt = '%Y/%m/%d'
    else:
        fmt = '%Y/%m/%d %-I:%M %p'
    return date_obj.astimezone().strftime(fmt)


def hbar(tl, widths=None, separator='╋'):
    if tl == 1:
        val = '┄'
    elif tl == 2:
        val = '┄┄'
    elif tl == 3:
        val = '┄┉┄'
    else:
        val = '┄┉' + '━' * (tl - 4) + '┉┄'

    if not widths:
        print(val)
        return

    ret = list(val)
    pos = 0
    for idx in range(0, len(widths) - 1):
        pos = pos + widths[idx] + 2
        ret[pos - 1] = separator
        pos = pos + 1
    print(''.join(ret))


def hbar_over(text):
    tl = len(text)
    if not tl:
        return
    hbar(tl)
    print(text)


def hbar_under(text):
    tl = len(text)
    if not tl:
        return
    print(text)
    hbar(tl)


def nym(arg, underscore='+-. !?;:\'",\t', remove='()[]{}?<>/='):
    '''
    This is for allowing case and quote flexibility for strings when
    searching dictionaries or other data sets based on user input (esp.
    from the command line) where the likelihood of key collisions is
    low. For example, if we want to search a dictionary, we'd check the
    nym of the value provided with the nym of the key to see if they
    match. This should not be used when likelihood of collisions is high.
    (Origin: Greek word meaning "name")

    Parameters:
        arg (string): A string to create the nym for
        underscore (string): A set of characters to replace with underscores
        remove (string): A set of characters to remove from the return value

    Returns:
        ret (string): A lower-case string with characters translated
                      or removed.
    '''
    if (ret := arg) not in ('', None):
        tr = str.maketrans(underscore, '_' * len(underscore), remove)
        ret = str(arg).lower().translate(tr) or '_'

    return ret


_pretty_print = PrettyPrinter(indent=4)


def pretty_print(obj):
    _pretty_print.pprint(obj)


# Print stuff in a format like so:
# field1 | field2 | field3....
#                 | field3-continued...
#                 | field3-continued-more...
# linesplit: None = trail off with '..', or separator
#            character (space, comma, etc.)
#
# vsep_print(arg1, screen_width, width1, arg2, width2, ... argN)
#
vseparator = '┃'


def vsep_print(linesplit=None, screen_width=0, *vals):
    sep = f' {vseparator} '

    fields = []
    widths = []
    consumed = 0

    if not screen_width:
        screen_width = termios.tcgetwinsize(sys.stdout)[1]
    args = list(vals)

    if not args:
        return None

    while len(args) >= 2:
        val = args.pop(0)
        if not isinstance(val, str):
            val = str(val)
        fields.append(val)
        widths.append(int(args.pop(0)))
    fields.append(args.pop(0))

    wn = sum(widths) + (3 * len(widths)) + 1
    if wn > screen_width:
        print(f'Screen too narrow ({wn} > {screen_width}')
        return 0

    #       field widths+ separators
    width = sum(widths) + (len(fields) - 1) * len(sep)
    for idx in range(0, len(widths)):
        print(fields[idx].ljust(widths[idx]) + sep, end='')
        consumed += widths[idx] + len(sep)

    last = fields.pop()
    if not isinstance(last, str):
        last = str(last)

    if len(last) <= (screen_width - width):
        print(last)
        return len(last) + width

    # Longer than remaining horizontal screen
    if not linesplit:
        print(last[:(screen_width - width - 1)] + '…')
        return screen_width

    max_chunk_len = screen_width - width
    chunks = last.split(linesplit)
    # Start on first line
    newline = False
    consumed = 0

    wrap_line = sep.join([' ' * width for width in widths]) + sep
    while len(chunks):
        chunk = chunks.pop(0)
        while len(chunk):
            if newline is True:
                consumed = 0
                print(wrap_line, end='')
            # Assume we'll get to the next line
            newline = True
            if len(chunk) > max_chunk_len:
                # If we've started a line, start a very long
                # max-chunk-len on new line
                if consumed > 0:
                    erase_line(True)
                    newline = True
                    continue
                print(chunk[:max_chunk_len])
                newline = True
                chunk = chunk[max_chunk_len:]
                continue
            if len(chunk) > (max_chunk_len - consumed):
                # Start this chunk on new line
                erase_line(True)
                newline = True
                continue
            # OK, we have space - print it
            print(chunk, end='')
            consumed += (len(chunk) + 1)
            if consumed < (max_chunk_len - 1):
                print(linesplit, end='')
                newline = False
            else:
                erase_line(True)
                newline = True
            # Next chunk
            break
    # If we terminate on the edge of the screen, we already
    # print newline above, so just a check to avoid an erroneous
    # newline in output
    if consumed < max_chunk_len and not newline:
        erase_line(True)
    return screen_width


def get_colors():
    if not fancy_output:
        return None
    if not color_shift:
        return None
    if not sys.stdin.isatty() or not sys.stdout.isatty():
        return None
    curr_attrs = termios.tcgetattr(sys.stdin)
    new_attrs = copy.deepcopy(curr_attrs)
    new_attrs[3] = curr_attrs[3] & ~(termios.ICANON | termios.ECHO)
    termios.tcsetattr(sys.stdin, termios.TCSAFLUSH, new_attrs)
    sys.stdout.write(']10;?')
    sys.stdout.flush()
    fg = ''
    while (fg := fg + sys.stdin.read(1)):
        if fg.endswith(''):
            break
    sys.stdout.write(']11;?')
    sys.stdout.flush()
    bg = ''
    while (bg := bg + sys.stdin.read(1)):
        if bg.endswith(''):
            break
    termios.tcsetattr(sys.stdin, termios.TCSAFLUSH, curr_attrs)

    # hack it up
    rx = r';rgb:(([0-9a-f]+)/([0-9a-f]+)/([0-9a-f]+))'
    fgm = re.findall(rx, fg)
    if not fgm:
        return None
    bgm = re.findall(rx, bg)
    if not bgm:
        return None

    # intify - we get a 16-bit hex from our terminal
    # query, but when we set, it's 8-bit / truecolor.
    # Let's hope this shifty stuff is enough.
    fgr = [int(val, 16) >> 8 for val in fgm[0][1:]]
    bgr = [int(val, 16) >> 8 for val in bgm[0][1:]]

    return [fgr, bgr]


def erase_line(newline=False):
    if newline:
        sys.stdout.write('\n')
    if fancy_output:
        sys.stdout.write('[K[2K')


def set_color(fg=None, bg=None, erase=False):
    if fg:
        sys.stdout.write(f'[38;2;{fg[0]};{fg[1]};{fg[2]}m')
    if bg:
        # When you set bgcolor, if you want the whole line to
        # have it, you have to erase the line before writing to
        # it - that's why we have the erase parameter.
        sys.stdout.write(f'[48;2;{bg[0]};{bg[1]};{bg[2]}m')
        if erase:
            erase_line()


def pretty_matrix(matrix, header=True, header_bar=True):
    global color_shift

    # Range test color_shift
    if not isinstance(color_shift, int):
        color_shift = 0
    if color_shift > 128 or color_shift < 0:
        color_shift = 0

    colors = None
    if colors := get_colors():
        bgcolor = colors[1]
        tint = [0] * len(bgcolor)
        for idx in range(0, len(bgcolor)):
            if bgcolor[idx] >= ((1 << 8) - color_shift):
                tint[idx] = bgcolor[idx] - color_shift
            else:
                tint[idx] = bgcolor[idx] + color_shift

    # total # of printed lines less header
    lines = 0
    even = False
    screen_width = termios.tcgetwinsize(sys.stdout)[1]
    # Renders a table with the right-most field truncated/wrapped if needed.
    # Undefined if the screen width is too wide to accommodate all but the
    # last field
    col_widths = len(matrix[0]) * [0]
    for row in matrix:
        if len(row) != len(col_widths):
            raise ValueError('Column count mismatch')
        for val in range(0, len(col_widths)):
            if isinstance(row[val], str):
                vlen = len(row[val])
            else:
                vlen = len(str(row[val]))
            col_widths[val] = max(col_widths[val], vlen)
    if header:
        start = 1
        line = []
        for item in range(0, len(col_widths)):
            line.extend([matrix[0][item], col_widths[item]])
        line.pop()
        # XXX Do we want to render full-width here?
        width = vsep_print(' ', screen_width, *line)
        if not width:
            return
    else:
        width = min(sum(col_widths) + 3 * len(col_widths) + 1, screen_width)
        start = 0
    if header_bar:
        if header:
            sep = '╋'
        else:
            sep = '┳'
        hbar(width, col_widths, sep)
    for row in matrix[start:]:
        line = []
        for item in range(0, len(col_widths)):
            line.extend([row[item], col_widths[item]])
        line.pop()
        if colors:
            if even:
                set_color(bg=tint, erase=True)
            else:
                set_color(bg=bgcolor, erase=True)
        even = not even

        vsep_print(' ', screen_width, *line)
        lines = lines + 1

    # Reset to default just in case
    if colors:
        set_color(bg=bgcolor, erase=True)

    return lines


def native_csv(matrix, header=True, header_bar=True):
    lines = 0
    # header_bar currently unused
    csv_out = csv.writer(sys.stdout)
    if header:
        start = 0
    else:
        start = 1
    for row in matrix[start:]:
        csv_out.writerow(row)
        lines = lines + 1

    return lines


_writers = {
    'default': pretty_matrix,
    'csv': native_csv
}


def render_matrix(matrix, header=True, header_bar=True, fmt='default'):
    if fmt not in _writers:
        fmt = 'default'
    return _writers[fmt](matrix, header, header_bar)



#!/usr/bin/python3
#
# Locked read/write ops for binary state files
#
import os
import struct
import fcntl
import pickle
import time


def lock(fd, block=True):
    lck = struct.pack('hhllhh', fcntl.F_WRLCK, 0, 0, 0, 0, 0)
    op = fcntl.F_SETLKW
    if block:
        op = fcntl.F_SETLK
    fcntl.fcntl(fd, op, lck)


def pickle_read(fn, **kwargs):
    ret = None
    exp = 0

    block = True
    if 'block' in kwargs and kwargs['block'] is False:
        block = False

    if 'expire' in kwargs:
        ex = kwargs['expire']
        if ex is None:
            exp = 0
        elif isinstance(ex, bool) and ex is False:
            exp = 0
        elif isinstance(ex, str) and ex == '0':
            exp = 0
        elif isinstance(ex, int) is int and ex <= 0:
            exp = 0
        elif isinstance(ex, float) and ex <= 0:
            exp = 0
        else:
            exp = ex

    inp = os.path.expanduser(fn)
    if not inp:
        return None

    try:
        st = os.stat(inp)
    except OSError:
        return None

    # Expired cache
    if exp and time.gmtime(st.st_mtime + float(exp)) <= time.gmtime():
        return None

    try:
        fd = os.open(inp, os.O_RDWR)
        lock(fd, block)
    except Exception as ex:  # NOQA - ditto
        if fd != -1:
            os.close(fd)
        return None

    fp = open(fd, 'rb')
    ret = pickle.load(fp)
    fp.close()
    return ret


def pickle_write(fn, obj):
    # TODO: mkstemp() and os.rename() for atomicity
    inp = os.path.expanduser(fn)
    if not inp:
        return None

    fd = -1
    try:
        fd = os.open(inp, os.O_WRONLY | os.O_SYNC | os.O_CREAT, mode=0o600)
        # lock before writing
        lock(fd)
    except:  # NOQA - Lots of reasons this could fail.
        if fd != -1:
            os.close(fd)
        return None

    os.lseek(fd, 0, 0)
    os.truncate(fd, 0)
    fp = open(fd, 'wb')
    ret = pickle.dump(obj, fp)
    fp.close()
    return ret



#!/usr/bin/env python

import jirate.jira_fields

import pytest


stringy_value = "haha"

singleton_value = {
    'name': 'TestValue',
    'key': 'thekey',
    'value': 'thevalue',
    'emailAddress': 'robert@thepie.com',
    'displayName': 'Rory Obert'
}

list_value = [
    'val1',
    'val2'
]

listy_field_values = [
    {'emailAddress': 'abc@def',
     'displayName': 'Aloha Brings Customers',
     'value': 'value1',
     'name': 'name1'},
    {'emailAddress': 'def@def',
     'displayName': 'Dark Existential Ferrets',
     'value': 'value2',
     'name': 'name'}
]


fields = {
    'string': stringy_value,
    'singleton': singleton_value,
    'list': list_value,
    'dict_list': listy_field_values
}


def test_parser_string():
    assert jirate.jira_fields.string(stringy_value, fields) == stringy_value


def test_parser_key():
    assert jirate.jira_fields.key(singleton_value, fields) == 'thekey'

    with pytest.raises(KeyError):
        jirate.jira_fields.key({}, fields)


def test_parser_value():
    assert jirate.jira_fields.value(singleton_value, fields) == 'thevalue'

    with pytest.raises(KeyError):
        jirate.jira_fields.value({}, fields)


def test_parser_name():
    assert jirate.jira_fields.name(singleton_value, fields) == 'TestValue'

    with pytest.raises(KeyError):
        jirate.jira_fields.name({}, fields)


def test_parser_user():
    assert jirate.jira_fields.user(singleton_value, fields) == 'Rory Obert - robert@thepie.com'

    with pytest.raises(KeyError):
        jirate.jira_fields.user({}, fields)



#!/usr/bin/env python3

from pathlib import Path
import yaml

import jsonschema
import pytest

SCHEMA_DIR = Path('jirate/schemas').absolute()
SCHEMAS = list(SCHEMA_DIR.rglob("*.yaml"))


def test_at_least_one_schema():
    # Detects issues in the provided schemas path, as pytest quietly skips
    # parametrized tests with empty parameters
    assert len(SCHEMAS) > 0


@pytest.mark.parametrize("schema", SCHEMAS)
def test_schema(schema):
    schema = yaml.safe_load(schema.read_text())
    jsonschema.Draft202012Validator.check_schema(schema)



#!/usr/bin/python3

from jirate.template_vars import apply_values

import pytest  # NOQA


def test_apply_var_type1():
    inp = """
{% set var=var or "1.0" %}
fork: {@var@}
"""
    exp = 'fork: 1.0'
    assert apply_values(inp).strip() == exp.strip()


def test_apply_var_type2():
    inp = """
fork: {@var|default('1.0')@}
"""
    exp = 'fork: 1.0'
    assert apply_values(inp).strip() == exp.strip()


def test_missing_var():
    inp = """
fork: {@var|default('1.0')@}
knife: {@butter@}
"""
    with pytest.raises(ValueError):
        apply_values(inp)


def test_provided_var():
    inp = """
fork: {@var|default('1.0')@}
knife: {@butter@}
"""
    exp = """
fork: 2.0
knife: salted
"""
    values = {'var': '2.0',
              'butter': 'salted'}
    assert apply_values(inp, values).strip() == exp.strip()


def test_invalid_var():
    inp = """
fork: {@var|default('1.0')@}
knife: {@butter@}
"""
    values = {'var': '2.0',
              'butter': 'salted',
              'bacon': 'such delishus, very salt, fat, wow'}
    with pytest.raises(ValueError):
        apply_values(inp, values)


def test_ignore_underscore():
    # Variables that start with underscore are
    # ignored and thus are not required to be set
    # during template rendering.
    inp = """
{% set _varone=_varone or "1.0" %}
fork: {@_vartwo|default('1.0')@}
knife: {@_varthree@}
spoon: {@varfour@}
"""

    apply_values(inp, {'varfour': 'meh'})


def test_underscore_invalid():
    # It's a user error to try to set a variable that
    # starts with an underscore
    inp = """
{% set _varone=_varone or "1.0" %}
fork: {@_vartwo|default('1.0')@}
knife: {@_varthree@}
spoon: {@varfour@}
"""

    with pytest.raises(ValueError):
        apply_values(inp, {'_varthree': 'abc', 'varfour': 'meh'})


def test_datetime_strftime():
    inp = """
{% set year = datetime(2025, 5, 7, 10, 32, 27, 87081).strftime('%Y') %}
year: {@year@}
"""
    exp = 'year: 2025'

    assert apply_values(inp).strip() == exp.strip()


def test_datetime_override():
    inp = """
{% set year = year or datetime(2025, 5, 7, 10, 32, 27, 87081).strftime('%Y') %}
year: {@year@}
"""
    exp = 'year: 2033'

    values = {'year': '2033'}

    assert apply_values(inp, values).strip() == exp.strip()
